//------------------------------------------------------------
// Copyright (c) Microsoft Corporation.  All rights reserved.
//------------------------------------------------------------

// This file isn't generated, but this comment is necessary to exclude it from StyleCop analysis.
// <auto-generated/>

using System;
using System.Linq;
using System.Text;

using static System.Text.Encoding;
using static Microsoft.Data.Encryption.Resources.Strings;

namespace Microsoft.Data.Encryption.Cryptography.Serializers
{
    /// <summary>
    /// Contains the methods for serializing and deserializing <see cref="string"/> type data objects
	/// that is compatible with the Always Encrypted feature in SQL Server and Azure SQL.
    /// </summary>
    internal sealed class SqlCharSerializer : SqlCodePageEncodingSerializer
    {
        private const int DefaultSize = 30;
        private const int MinSize = 1;
        private const int MaxSize = 8000;
        private static readonly char Padding = ' ';

        /// <summary>
        /// The default character encoding Windows-1252. It is also referred to as "ANSI".
        /// </summary>
        private const int DefaultEncodingCodePoint = 1252;

        /// <summary>
        /// The <see cref="Identifier"/> uniquely identifies a particular Serializer implementation.
        /// </summary>
        public override string Identifier => "SQL_Char_Nullable";

        private int size;

        /// <summary>
        /// Gets or sets the maximum length of the data.
        /// </summary>
        /// <remarks>
        /// If not explicitly set, the size is defaulted to 30.
        /// </remarks>
        /// <exception cref="MicrosoftDataEncryptionException">
        /// Thrown when set to a value that is out of the valid range [1 - 8000] for this setting.
        /// </exception>
        public int Size
        {
            get => size;
            set
            {
                if (value < MinSize || value > MaxSize)
                {
                    throw new MicrosoftDataEncryptionException(ValueOutOfRange.Format(value));
                }

                size = value;
            }
        }

        private Encoding characterEncoding;

        /// <summary>
        /// Gets or sets the character encoding.
        /// </summary>
        /// <remarks>
        /// If not explicitly set, the encoding is defaulted to Windows-1252, which is also referred to as "ANSI".
        /// </remarks>
        public int CodePageCharacterEncoding
        {
            get => characterEncoding.CodePage;
            set => characterEncoding = GetEncoding(value);
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SqlCharSerializer"/> class.
        /// </summary>
        /// <param name="size">The maximum length of the data</param>
        /// <param name="codePageCharacterEncoding">The code page character encoding.</param>
        /// <exception cref="MicrosoftDataEncryptionException">
        /// Thrown when set to a value that is out of the valid range [1 - 8000] for this setting.
        /// </exception>
        public SqlCharSerializer(int size = DefaultSize, int codePageCharacterEncoding = DefaultEncodingCodePoint)
        {
            Size = size;
            characterEncoding = GetEncoding(codePageCharacterEncoding);
        }

        /// <summary>
        /// Deserializes the provided <paramref name="bytes"/>
        /// </summary>
        /// <param name="bytes">The data to be deserialized</param>
        /// <returns>The serialized data</returns>
        /// <remarks>
        /// If the deserialized value's length is less than the size, then the value will be padded on the
        /// left to match the size. Padding is achieved by using spaces.
        /// </remarks>
        public override string Deserialize(byte[] bytes)
        {
            return bytes.IsNull() ? null : PadToLength(characterEncoding.GetString(bytes));
        }

        /// <summary>
        /// Serializes the provided <paramref name="value"/>
        /// </summary>
        /// <param name="value">The value to be serialized</param>
        /// <returns>The serialized data as a byte array</returns>
        /// <remarks>
        /// If the value's length is greater than the size, then the value will be truncated on the left to match the size.
        /// </remarks>
        public override byte[] Serialize(string value)
        {
            return value.IsNull() ? null : characterEncoding.GetBytes(TrimToLength(value));
        }

        private string TrimToLength(string value) => value.Length > Size ? new string(value.Take(Size).ToArray()) : value;

        private string PadToLength(string value) => value.Length < Size ? value.PadRight(Size, Padding) : value;
    }
}
