﻿//------------------------------------------------------------
// Copyright (c) Microsoft Corporation.  All rights reserved.
//------------------------------------------------------------

namespace Microsoft.Azure.Cosmos.SDK.EmulatorTests
{
    using System;
    using System.Collections.Generic;
    using System.Net;
    using System.Threading.Tasks;
    using Microsoft.Azure.Cosmos.EmulatorTests.Query;
    using Microsoft.Azure.Cosmos.Utils;
    using Microsoft.VisualStudio.TestTools.UnitTesting;

    [TestClass]
    public class CosmosPermissionTests : BaseCosmosClientHelper
    {
        [TestInitialize]
        public async Task TestInitialize()
        {
            await base.TestInit();
        }

        [TestCleanup]
        public async Task Cleanup()
        {
            await base.TestCleanup();
        }

        [TestMethod]
        public async Task CRUDTest()
        {
            string containerId = Guid.NewGuid().ToString();
            ContainerResponse containerResponse = await this.database.CreateContainerAsync(containerId, "/id");
            Assert.AreEqual(HttpStatusCode.Created, containerResponse.StatusCode);

            string userId = Guid.NewGuid().ToString();
            UserResponse userResponse = await this.database.CreateUserAsync(userId);
            User user = userResponse.User;
            Assert.AreEqual(HttpStatusCode.Created, userResponse.StatusCode);
            Assert.AreEqual(userId, user.Id);

            string permissionId = Guid.NewGuid().ToString();
            PermissionProperties permissionProperties = new PermissionProperties(permissionId, PermissionMode.Read, containerResponse.Container);
            PermissionResponse permissionResponse = await user.CreatePermissionAsync(permissionProperties);
            Assert.AreEqual(HttpStatusCode.Created, userResponse.StatusCode);
            Assert.AreEqual(permissionId, permissionResponse.Resource.Id);
            Assert.AreEqual(permissionProperties.PermissionMode, permissionResponse.Resource.PermissionMode);
            Assert.IsNotNull(permissionResponse.Resource.Token);
            SelflinkValidator.ValidatePermissionSelfLink(permissionResponse.Resource.SelfLink);

            PermissionProperties newPermissionProperties = new PermissionProperties(permissionId, PermissionMode.All, containerResponse.Container);
            permissionResponse = await user.GetPermission(permissionId).ReplaceAsync(newPermissionProperties);
            //Backend returns Created instead of OK
            Assert.AreEqual(HttpStatusCode.Created, userResponse.StatusCode);
            Assert.AreEqual(permissionId, permissionResponse.Resource.Id);
            Assert.AreEqual(newPermissionProperties.PermissionMode, permissionResponse.Resource.PermissionMode);
            SelflinkValidator.ValidatePermissionSelfLink(permissionResponse.Resource.SelfLink);

            permissionResponse = await user.GetPermission(permissionId).ReadAsync();
            Assert.AreEqual(HttpStatusCode.OK, permissionResponse.StatusCode);
            Assert.AreEqual(permissionId, permissionResponse.Resource.Id);
            SelflinkValidator.ValidatePermissionSelfLink(permissionResponse.Resource.SelfLink);

            permissionResponse = await user.GetPermission(permissionId).DeleteAsync();
            Assert.AreEqual(HttpStatusCode.NoContent, permissionResponse.StatusCode);

            try
            {
                permissionResponse = await user.GetPermission(permissionId).ReadAsync();
                Assert.Fail();
            }
            catch(CosmosException ex)
            {
                Assert.AreEqual(HttpStatusCode.NotFound, ex.StatusCode);
            }

            permissionId = Guid.NewGuid().ToString();
            permissionProperties = new PermissionProperties(permissionId, PermissionMode.Read, containerResponse.Container);
            permissionResponse = await user.CreatePermissionAsync(permissionProperties);
            Assert.AreEqual(HttpStatusCode.Created, userResponse.StatusCode);
            Assert.AreEqual(permissionId, permissionResponse.Resource.Id);
            Assert.AreEqual(permissionProperties.PermissionMode, permissionResponse.Resource.PermissionMode);
            Assert.IsNotNull(permissionResponse.Resource.Token);

            newPermissionProperties = new PermissionProperties(permissionId, PermissionMode.All, containerResponse.Container);
            permissionResponse = await user.UpsertPermissionAsync(newPermissionProperties);
            Assert.AreEqual(HttpStatusCode.OK, permissionResponse.StatusCode);
            Assert.AreEqual(permissionId, permissionResponse.Resource.Id);
            Assert.AreEqual(newPermissionProperties.PermissionMode, permissionResponse.Resource.PermissionMode);
        }

        [TestMethod]
        [DataRow(ConnectionMode.Gateway)]
        [DataRow(ConnectionMode.Direct)]
        public async Task ContainerResourcePermissionTest(ConnectionMode mode)
        {
            CosmosClientOptions cosmosClientOptions = new CosmosClientOptions()
            {
                ConnectionMode = mode,
            };

            //create user
            string userId = Guid.NewGuid().ToString();
            UserResponse userResponse = await this.database.CreateUserAsync(userId);
            Assert.AreEqual(HttpStatusCode.Created, userResponse.StatusCode);
            Assert.AreEqual(userId, userResponse.Resource.Id);
            User user = userResponse.User;

            //create resource
            string containerId = Guid.NewGuid().ToString();
            ContainerResponse containerResponse = await this.database.CreateContainerAsync(containerId, "/id");
            Assert.AreEqual(HttpStatusCode.Created, containerResponse.StatusCode);
            Container container = containerResponse.Container;

            //create permission
            string permissionId = Guid.NewGuid().ToString();
            PermissionProperties permissionProperties = new PermissionProperties(permissionId, PermissionMode.Read, container);
            PermissionResponse permissionResponse = await user.CreatePermissionAsync(permissionProperties);
            PermissionProperties permission = permissionResponse.Resource;
            Assert.AreEqual(HttpStatusCode.Created, userResponse.StatusCode);
            Assert.AreEqual(permissionId, permission.Id);
            Assert.AreEqual(permissionProperties.PermissionMode, permission.PermissionMode);

            using (CosmosClient tokenCosmosClient = TestCommon.CreateCosmosClient(cosmosClientOptions, resourceToken: permission.Token))
            {
                Container readContainerRef = tokenCosmosClient.GetContainer(this.database.Id, containerId);

                //read resource with PermissionMode.Read
                using FeedIterator<dynamic> feedIterator = readContainerRef.GetItemQueryIterator<dynamic>("SELECT * FROM c");
                while (feedIterator.HasMoreResults)
                {
                    _  = await feedIterator.ReadNextAsync();
                }

                //delete resource with PermissionMode.Read
                try
                {
                    ContainerResponse response = await tokenCosmosClient
                    .GetDatabase(this.database.Id)
                    .GetContainer(containerId)
                    .DeleteContainerAsync();
                    Assert.Fail();
                }
                catch (CosmosException ex)
                {
                    Assert.AreEqual(HttpStatusCode.Forbidden, ex.StatusCode);
                }
            }

            //update permission to PermissionMode.All
            permissionProperties = new PermissionProperties(permissionId, PermissionMode.All, container);
            permissionResponse = await user.GetPermission(permissionId).ReplaceAsync(permissionProperties);
            permission = permissionResponse.Resource;

            //delete resource with PermissionMode.All
            using (CosmosClient tokenCosmosClient = TestCommon.CreateCosmosClient(cosmosClientOptions, resourceToken: permission.Token))
            {
                ContainerResponse response = await tokenCosmosClient
                    .GetDatabase(this.database.Id)
                    .GetContainer(containerId)
                    .DeleteContainerAsync();
                Assert.AreEqual(HttpStatusCode.NoContent, response.StatusCode);
            }
        }

        [TestMethod]
        [DataRow(ConnectionMode.Gateway)]
        [DataRow(ConnectionMode.Direct)]
        public async Task ContainerPartitionResourcePermissionTest(ConnectionMode connectionMode)
        {
            CosmosClientOptions cosmosClientOptions = new CosmosClientOptions()
            {
                ConnectionMode = connectionMode
            };

            using CosmosClient cosmosClient = TestCommon.CreateCosmosClient(cosmosClientOptions);

            Database database = await cosmosClient.CreateDatabaseIfNotExistsAsync("PermissionTest");

            //create user
            string userId = Guid.NewGuid().ToString();
            UserResponse userResponse = await database.CreateUserAsync(userId);
            Assert.AreEqual(HttpStatusCode.Created, userResponse.StatusCode);
            Assert.AreEqual(userId, userResponse.Resource.Id);
            User user = userResponse.User;

            //create resource
            string containerId = Guid.NewGuid().ToString();

            ContainerResponse containerResponse = await database.CreateContainerAsync(
                id: containerId,
                partitionKeyPath: "/id");

            Assert.AreEqual(HttpStatusCode.Created, containerResponse.StatusCode);
            Container container = containerResponse.Container;

            // Create items to read
            ToDoActivity itemAccess = ToDoActivity.CreateRandomToDoActivity();
            ToDoActivity itemNoAccess = ToDoActivity.CreateRandomToDoActivity();

            await container.CreateItemAsync<ToDoActivity>(
                itemAccess,
                new PartitionKey(itemAccess.id));

            await container.CreateItemAsync<ToDoActivity>(
                itemNoAccess,
                new PartitionKey(itemNoAccess.id));

            //create permission
            string permissionId = Guid.NewGuid().ToString();
            PartitionKey partitionKey = new PartitionKey(itemAccess.id);
            PermissionProperties permissionProperties = new PermissionProperties(
                permissionId,
                PermissionMode.Read,
                container,
                partitionKey);

            PermissionResponse permissionResponse = await user.CreatePermissionAsync(permissionProperties);
            PermissionProperties permission = permissionResponse.Resource;
            Assert.AreEqual(HttpStatusCode.Created, userResponse.StatusCode);
            Assert.AreEqual(permissionId, permission.Id);
            Assert.AreEqual(permissionProperties.PermissionMode, permission.PermissionMode);

            using (CosmosClient tokenCosmosClient = TestCommon.CreateCosmosClient(clientOptions: cosmosClientOptions, resourceToken: permission.Token))
            {
                Container tokenContainer = tokenCosmosClient.GetContainer(database.Id, containerId);
                await tokenContainer.ReadItemAsync<ToDoActivity>(itemAccess.id, new PartitionKey(itemAccess.id));

                try
                {
                    await tokenContainer.ReadItemAsync<ToDoActivity>(itemNoAccess.id, new PartitionKey(itemNoAccess.id));
                    Assert.Fail();
                }
                catch (CosmosException ex)
                {
                    Assert.AreEqual(HttpStatusCode.Forbidden, ex.StatusCode);
                }

                QueryRequestOptions queryRequestOptions = new QueryRequestOptions()
                {
                    PartitionKey = new PartitionKey(itemAccess.id)
                };

                FeedIterator<ToDoActivity> feedIterator = tokenContainer.GetItemQueryIterator<ToDoActivity>(
                    queryText: "select * from T",
                    requestOptions: queryRequestOptions);

                List<ToDoActivity> result = new List<ToDoActivity>();
                while (feedIterator.HasMoreResults)
                {
                    FeedResponse<ToDoActivity> toDoActivities = await feedIterator.ReadNextAsync();
                    result.AddRange(toDoActivities);
                }

                Assert.AreEqual(1, result.Count);

                // Test query with no service interop via gateway query plan to replicate x32 app
                ContainerInternal containerCore = (ContainerInlineCore)tokenContainer;
                MockCosmosQueryClient mock = new MockCosmosQueryClient(
                    clientContext: containerCore.ClientContext,
                    cosmosContainerCore: containerCore,
                    forceQueryPlanGatewayElseServiceInterop: true);

                Container tokenGatewayQueryPlan = new ContainerInlineCore(
                    containerCore.ClientContext,
                    (DatabaseInternal)containerCore.Database,
                    containerCore.Id,
                    mock);

                FeedIterator<ToDoActivity> feedIteratorGateway = tokenGatewayQueryPlan.GetItemQueryIterator<ToDoActivity>(
                    queryText: "select * from T",
                    requestOptions: queryRequestOptions);

                List<ToDoActivity> resultGateway = new List<ToDoActivity>();
                while (feedIteratorGateway.HasMoreResults)
                {
                    FeedResponse<ToDoActivity> toDoActivities = await feedIteratorGateway.ReadNextAsync();
                    resultGateway.AddRange(toDoActivities);
                }

                Assert.AreEqual(1, resultGateway.Count);
            }

            await database.DeleteStreamAsync();
        }

        [TestMethod]
        [DataRow(ConnectionMode.Gateway)]
        [DataRow(ConnectionMode.Direct)]
        public async Task ItemResourcePermissionTest(ConnectionMode connectionMode)
        {
            CosmosClientOptions cosmosClientOptions = new CosmosClientOptions()
            {
                ConnectionMode = connectionMode
            };

            //create user
            string userId = Guid.NewGuid().ToString();
            UserResponse userResponse = await this.database.CreateUserAsync(userId);
            Assert.AreEqual(HttpStatusCode.Created, userResponse.StatusCode);
            Assert.AreEqual(userId, userResponse.Resource.Id);
            User user = userResponse.User;

            //create resource
            string containerId = Guid.NewGuid().ToString();
            ContainerResponse containerResponse = await this.database.CreateContainerAsync(containerId, "/id");
            Assert.AreEqual(HttpStatusCode.Created, containerResponse.StatusCode);
            Container container = containerResponse.Container;
            string itemId = Guid.NewGuid().ToString();
            PartitionKey partitionKey = new PartitionKey(itemId);
            ItemResponse<dynamic> itemRespnose = await container.CreateItemAsync<dynamic>(new { id = itemId }, partitionKey);
            Assert.AreEqual(HttpStatusCode.Created, itemRespnose.StatusCode);

            //create permission
            string permissionId = Guid.NewGuid().ToString();
            PermissionProperties permissionProperties = new PermissionProperties(permissionId, PermissionMode.Read, container, partitionKey, itemId);
            Assert.AreEqual(permissionProperties.ResourceUri, $"dbs/{container.Database.Id}/colls/{container.Id}/docs/{itemId}");
            PermissionResponse permissionResponse = await user.CreatePermissionAsync(permissionProperties);
            PermissionProperties permission = permissionResponse.Resource;
            Assert.AreEqual(HttpStatusCode.Created, userResponse.StatusCode);
            Assert.AreEqual(permissionId, permission.Id);
            Assert.AreEqual(permissionProperties.PermissionMode, permission.PermissionMode);

            using (CosmosClient tokenCosmosClient = TestCommon.CreateCosmosClient(clientOptions: cosmosClientOptions, resourceToken: permission.Token))
            {
                Container tokenContainer = tokenCosmosClient.GetContainer(this.database.Id, containerId);

                //read resource with PermissionMode.Read
                ItemResponse<dynamic> readPermissionItem = await tokenContainer.ReadItemAsync<dynamic>(itemId, partitionKey);
                Assert.AreEqual(itemId, readPermissionItem.Resource.id.ToString());

                //delete resource with PermissionMode.Read
                try
                {
                    ItemResponse<dynamic> response = await tokenContainer.DeleteItemAsync<dynamic>(
                        itemId,
                        partitionKey);

                    Assert.Fail();
                }
                catch (CosmosException ex)
                {
                    Assert.AreEqual(HttpStatusCode.Forbidden, ex.StatusCode);
                }
            }

            //update permission to PermissionMode.All
            permissionProperties = new PermissionProperties(permissionId, PermissionMode.All, container);
            permissionResponse = await user.GetPermission(permissionId).ReplaceAsync(permissionProperties);
            permission = permissionResponse.Resource;

            //delete resource with PermissionMode.All
            using (CosmosClient tokenCosmosClient = TestCommon.CreateCosmosClient(clientOptions: cosmosClientOptions, resourceToken: permission.Token))
            {
                using (FeedIterator<dynamic> feed = tokenCosmosClient
                    .GetDatabase(this.database.Id)
                    .GetContainer(containerId)
                    .GetItemQueryIterator<dynamic>(new QueryDefinition("select * from t")))
                {
                    while (feed.HasMoreResults)
                    {
                        FeedResponse<dynamic> response = await feed.ReadNextAsync();
                        Assert.IsNotNull(response);
                    }
                }
            }
        }

        [TestMethod]
        [DataRow(ConnectionMode.Gateway)]
        [DataRow(ConnectionMode.Direct)]
        public async Task EnsureUnauthorized_ThrowsCosmosClientException(ConnectionMode connectionMode)
        {
            CosmosClientOptions cosmosClientOptions = new CosmosClientOptions()
            {
                ConnectionMode = connectionMode
            };

            string authKey = ConfigurationManager.AppSettings["MasterKey"];
            string endpoint = ConfigurationManager.AppSettings["GatewayEndpoint"];

            // Take the key and change some middle character
            authKey = authKey.Replace("m", "M");

            using CosmosClient cosmosClient = new CosmosClient(
                endpoint,
                authKey,
                cosmosClientOptions);

            CosmosException exception = await Assert.ThrowsExceptionAsync<CosmosException>(() => cosmosClient.GetContainer("test", "test").ReadItemAsync<dynamic>("test", new PartitionKey("test")));
            Assert.AreEqual(HttpStatusCode.Unauthorized, exception.StatusCode);
        }

        [TestMethod]
        [DataRow(ConnectionMode.Gateway)]
        [DataRow(ConnectionMode.Direct)]
        public async Task EnsureUnauthorized_ThrowsCosmosClientException_ReadAccountAsync(ConnectionMode connectionMode)
        {
            CosmosClientOptions cosmosClientOptions = new CosmosClientOptions()
            {
                ConnectionMode = connectionMode
            };

            string authKey = ConfigurationManager.AppSettings["MasterKey"];
            string endpoint = ConfigurationManager.AppSettings["GatewayEndpoint"];

            // Take the key and change some middle character
            authKey = authKey.Replace("m", "M");
            using CosmosClient  cosmosClient = new CosmosClient(
                endpoint,
                authKey,
                cosmosClientOptions);

            CosmosException exception1 = await Assert.ThrowsExceptionAsync<CosmosException>(() => cosmosClient.ReadAccountAsync());
            Assert.AreEqual(HttpStatusCode.Unauthorized, exception1.StatusCode);

        }

        [TestMethod]
        [DataRow(ConnectionMode.Gateway)]
        [DataRow(ConnectionMode.Direct)]
        public async Task EnsureUnauthorized_Writes_ThrowsCosmosClientException(ConnectionMode connectionMode)
        {
            CosmosClientOptions cosmosClientOptions = new CosmosClientOptions()
            {
                ConnectionMode = connectionMode
            };

            string authKey = ConfigurationManager.AppSettings["MasterKey"];
            string endpoint = ConfigurationManager.AppSettings["GatewayEndpoint"];

            // Take the key and change some middle character
            authKey = authKey.Replace("m", "M");

            using CosmosClient cosmosClient = new CosmosClient(
                endpoint,
                authKey,
                cosmosClientOptions);

            CosmosException exception = await Assert.ThrowsExceptionAsync<CosmosException>(() => cosmosClient.GetContainer("test", "test").CreateItemAsync<dynamic>(new { id = "test" }));
            Assert.AreEqual(HttpStatusCode.Unauthorized, exception.StatusCode);
        }

        [TestMethod]
        [DataRow(ConnectionMode.Gateway)]
        [DataRow(ConnectionMode.Direct)]
        public async Task EnsureUnauthorized_Query_ThrowsCosmosClientException(ConnectionMode connectionMode)
        {
            CosmosClientOptions cosmosClientOptions = new CosmosClientOptions()
            {
                ConnectionMode = connectionMode
            };

            string authKey = ConfigurationManager.AppSettings["MasterKey"];
            string endpoint = ConfigurationManager.AppSettings["GatewayEndpoint"];

            // Take the key and change some middle character
            authKey = authKey.Replace("m", "M");

            using CosmosClient cosmosClient = new CosmosClient(
                endpoint,
                authKey,
                cosmosClientOptions);

            using FeedIterator<dynamic> iterator = cosmosClient.GetContainer("test", "test").GetItemQueryIterator<dynamic>("SELECT * FROM c");

            CosmosException exception = await Assert.ThrowsExceptionAsync<CosmosException>(() => iterator.ReadNextAsync());
            Assert.AreEqual(HttpStatusCode.Unauthorized, exception.StatusCode);
        }
    }
}
